@@ -60,8 +60,11 @@ Guacamole.Keyboard = function(element) {
      * 
      * @event
      * @param {Number} keysym The keysym of the key being pressed.
+     * @returns {Boolean} true if the originating event of this keypress should
+     *                    be allowed through to the browser, false or undefined
+     *                    otherwise.
      */
-	this.onkeydown = null;
+    this.onkeydown = null;
 
     /**
      * Fired whenever the user releases a key with the element associated
@@ -69,8 +72,11 @@ Guacamole.Keyboard = function(element) {
      * 
      * @event
      * @param {Number} keysym The keysym of the key being released.
+     * @returns {Boolean} true if the originating event of this key release 
+     *                    should be allowed through to the browser, false or
+     *                    undefined otherwise.
      */
-	this.onkeyup = null;
+    this.onkeyup = null;
 
     /**
      * Map of known JavaScript keycodes which do not map to typable characters
@@ -127,10 +133,35 @@ Guacamole.Keyboard = function(element) {
         18:  0xFFE7  // alt
     };
 
-	// Single key state/modifier buffer
-	var modShift = false;
-	var modCtrl = false;
-	var modAlt = false;
+    /**
+     * All modifiers and their states.
+     */
+    this.modifiers = {
+        
+        /**
+         * Whether shift is currently pressed.
+         */
+        "shift": false,
+        
+        /**
+         * Whether ctrl is currently pressed.
+         */
+        "ctrl" : false,
+        
+        /**
+         * Whether alt is currently pressed.
+         */
+        "alt"  : false
+
+    };
+
+    /**
+     * The state of every key, indexed by keysym. If a particular key is
+     * pressed, the value of pressed for that keysym will be true. If a key
+     * is not currently pressed, the value for that keysym may be false or
+     * undefined.
+     */
+    this.pressed = [];
 
     var keydownChar = new Array();
 
@@ -138,19 +169,19 @@ Guacamole.Keyboard = function(element) {
     var repeatKeyTimeoutId = -1;
     var repeatKeyIntervalId = -1;
 
-	// Starts repeating keystrokes
-	function startRepeat(keySym) {
-		repeatKeyIntervalId = setInterval(function() {
+    // Starts repeating keystrokes
+    function startRepeat(keySym) {
+        repeatKeyIntervalId = setInterval(function() {
             sendKeyReleased(keySym);
             sendKeyPressed(keySym);
         }, 50);
-	}
+    }
 
-	// Stops repeating keystrokes
-	function stopRepeat() {
-		if (repeatKeyTimeoutId != -1) clearInterval(repeatKeyTimeoutId);
-		if (repeatKeyIntervalId != -1) clearInterval(repeatKeyIntervalId);
-	}
+    // Stops repeating keystrokes
+    function stopRepeat() {
+        if (repeatKeyTimeoutId != -1) clearInterval(repeatKeyTimeoutId);
+        if (repeatKeyIntervalId != -1) clearInterval(repeatKeyIntervalId);
+    }
 
 
     function getKeySymFromKeyIdentifier(shifted, keyIdentifier) {
@@ -194,8 +225,8 @@ Guacamole.Keyboard = function(element) {
     function getKeySymFromKeyCode(keyCode) {
 
         var keysym = null;
-		if (!modShift) keysym = unshiftedKeySym[keyCode];
-		else {
+        if (!guac_keyboard.modifiers.shift) keysym = unshiftedKeySym[keyCode];
+        else {
             keysym = shiftedKeySym[keyCode];
             if (keysym == null) keysym = unshiftedKeySym[keyCode];
         }
@@ -205,17 +236,33 @@ Guacamole.Keyboard = function(element) {
     }
 
 
-	// Sends a single keystroke over the network
-	function sendKeyPressed(keysym) {
-		if (keysym != null && guac_keyboard.onkeydown)
-			guac_keyboard.onkeydown(keysym);
-	}
+    // Sends a single keystroke over the network
+    function sendKeyPressed(keysym) {
 
-	// Sends a single keystroke over the network
-	function sendKeyReleased(keysym) {
-		if (keysym != null && guac_keyboard.onkeyup)
-			guac_keyboard.onkeyup(keysym);
-	}
+        // Mark key as pressed
+        guac_keyboard.pressed[keysym] = true;
+
+        // Send key event
+        if (keysym != null && guac_keyboard.onkeydown)
+            return guac_keyboard.onkeydown(keysym) != false;
+        
+        return true;
+
+    }
+
+    // Sends a single keystroke over the network
+    function sendKeyReleased(keysym) {
+
+        // Mark key as released
+        guac_keyboard.pressed[keysym] = false;
+
+        // Send key event
+        if (keysym != null && guac_keyboard.onkeyup)
+            return guac_keyboard.onkeyup(keysym) != false;
+
+        return true;
+
+    }
 
 
     var KEYDOWN = 1;
@@ -223,44 +270,41 @@ Guacamole.Keyboard = function(element) {
 
     var keySymSource = null;
 
-	// When key pressed
+    // When key pressed
     var keydownCode = null;
-	element.onkeydown = function(e) {
+    element.onkeydown = function(e) {
 
         // Only intercept if handler set
         if (!guac_keyboard.onkeydown) return true;
 
-		var keynum;
-		if (window.event) keynum = window.event.keyCode;
-		else if (e.which) keynum = e.which;
+        var keynum;
+        if (window.event) keynum = window.event.keyCode;
+        else if (e.which) keynum = e.which;
 
-		// Ctrl/Alt/Shift
-		if (keynum == 16)
-			modShift = true;
-		else if (keynum == 17)
-			modCtrl = true;
-		else if (keynum == 18)
-			modAlt = true;
+        // Ctrl/Alt/Shift
+        if (keynum == 16)      guac_keyboard.modifiers.shift = true;
+        else if (keynum == 17) guac_keyboard.modifiers.ctrl  = true;
+        else if (keynum == 18) guac_keyboard.modifiers.alt   = true;
 
+        // If keysym is defined for given key code, key events can come from
+        // KEYDOWN.
         var keysym = getKeySymFromKeyCode(keynum);
-        if (keysym) {
-            // Get keysyms and events from KEYDOWN
+        if (keysym)
             keySymSource = KEYDOWN;
-        }
 
-        // If modifier keys are held down, and we have keyIdentifier
-        else if ((modCtrl || modAlt) && e.keyIdentifier) {
+        // Otherwise, if modifier keys are held down, try to get from keyIdentifier
+        else if ((guac_keyboard.modifiers.ctrl || guac_keyboard.modifiers.alt) && e.keyIdentifier) {
 
             // Get keysym from keyIdentifier
-            keysym = getKeySymFromKeyIdentifier(modShift, e.keyIdentifier);
+            keysym = getKeySymFromKeyIdentifier(guac_keyboard.modifiers.shift, e.keyIdentifier);
 
             // Get keysyms and events from KEYDOWN
             keySymSource = KEYDOWN;
 
         }
 
+        // Otherwise, resort to KEYPRESS
         else
-            // Get keysyms and events from KEYPRESS
             keySymSource = KEYPRESS;
 
         keydownCode = keynum;
@@ -273,7 +317,7 @@ Guacamole.Keyboard = function(element) {
 
                 // Send event
                 keydownChar[keynum] = keysym;
-                sendKeyPressed(keysym);
+                var returnValue = sendKeyPressed(keysym);
 
                 // Clear old key repeat, if any.
                 stopRepeat();
@@ -281,16 +325,23 @@ Guacamole.Keyboard = function(element) {
                 // Start repeating (if not a modifier key) after a short delay
                 if (keynum != 16 && keynum != 17 && keynum != 18)
                     repeatKeyTimeoutId = setTimeout(function() { startRepeat(keysym); }, 500);
+
+                // Use return code provided by handler
+                return returnValue;
+
             }
 
+            // Default to canceling event if no keypress is being sent, but
+            // source of events is keydown.
             return false;
+
         }
 
         return true;
 
-	};
+    };
 
-	// When key pressed
+    // When key pressed
     element.onkeypress = function(e) {
 
         // Only intercept if handler set
@@ -298,9 +349,9 @@ Guacamole.Keyboard = function(element) {
 
         if (keySymSource != KEYPRESS) return false;
 
-		var keynum;
-		if (window.event) keynum = window.event.keyCode;
-		else if (e.which) keynum = e.which;
+        var keynum;
+        if (window.event) keynum = window.event.keyCode;
+        else if (e.which) keynum = e.which;
 
         var keysym = getKeySymFromCharCode(keynum);
         if (keysym && keydownChar[keynum] != keysym) {
@@ -316,32 +367,34 @@ Guacamole.Keyboard = function(element) {
             stopRepeat();
 
             // Send key event
-            sendKeyPressed(keysym);
+            var returnValue = sendKeyPressed(keysym);
 
             // Start repeating (if not a modifier key) after a short delay
             repeatKeyTimeoutId = setTimeout(function() { startRepeat(keysym); }, 500);
+
+            return returnValue;
         }
 
+        // Default to canceling event if no keypress is being sent, but
+        // source of events is keypress.
         return false;
-	};
 
-	// When key released
-	element.onkeyup = function(e) {
+    };
+
+    // When key released
+    element.onkeyup = function(e) {
 
         // Only intercept if handler set
         if (!guac_keyboard.onkeyup) return true;
 
-		var keynum;
-		if (window.event) keynum = window.event.keyCode;
-		else if (e.which) keynum = e.which;
-		
-		// Ctrl/Alt/Shift
-		if (keynum == 16)
-			modShift = false;
-		else if (keynum == 17)
-			modCtrl = false;
-		else if (keynum == 18)
-			modAlt = false;
+        var keynum;
+        if (window.event) keynum = window.event.keyCode;
+        else if (e.which) keynum = e.which;
+        
+        // Ctrl/Alt/Shift
+        if (keynum == 16)      guac_keyboard.modifiers.shift = false;
+        else if (keynum == 17) guac_keyboard.modifiers.ctrl  = false;
+        else if (keynum == 18) guac_keyboard.modifiers.alt   = false;
         else
             stopRepeat();
 
@@ -352,16 +405,15 @@ Guacamole.Keyboard = function(element) {
         keydownChar[keynum] = null;
 
         // Send release event
-        sendKeyReleased(lastKeyDownChar);
+        return sendKeyReleased(lastKeyDownChar);
 
-		return false;
-	};
+    };
 
-	// When focus is lost, clear modifiers.
-	element.onblur = function() {
-		modAlt = false;
-		modCtrl = false;
-		modShift = false;
-	};
+    // When focus is lost, clear modifiers.
+    element.onblur = function() {
+        guac_keyboard.modifiers.alt = false;
+        guac_keyboard.modifiers.ctrl = false;
+        guac_keyboard.modifiers.shift = false;
+    };
 
 };
