@@ -44,8 +44,8 @@ var Guacamole = Guacamole || {};
  * mouse events into a non-browser-specific event provided by the
  * Guacamole.Mouse instance.
  * 
- * Touch event support is planned, but currently only in testing (translate
- * touch events into mouse events).
+ * Touch events are translated into mouse events as if the touches occurred
+ * on a touchpad (drag to push the mouse pointer, tap to click).
  * 
  * @constructor
  * @param {Element} element The Element to use to provide mouse events.
@@ -58,6 +58,24 @@ Guacamole.Mouse = function(element) {
      */
     var guac_mouse = this;
 
+    /**
+     * The distance a two-finger touch must move per scrollwheel event, in
+     * pixels.
+     */
+    this.scrollThreshold = 20 * (window.devicePixelRatio || 1);
+
+    /**
+     * The maximum number of milliseconds to wait for a touch to end for the
+     * gesture to be considered a click.
+     */
+    this.clickTimingThreshold = 250;
+
+    /**
+     * The maximum number of pixels to allow a touch to move for the gesture to
+     * be considered a click.
+     */
+    this.clickMoveThreshold = 10 * (window.devicePixelRatio || 1);
+
     /**
      * The current mouse state. The properties of this state are updated when
      * mouse events fire. This state object is also passed in as a parameter to
@@ -97,21 +115,33 @@ Guacamole.Mouse = function(element) {
      */
 	this.onmousemove = null;
 
-    function moveMouse(pageX, pageY) {
+    function cancelEvent(e) {
+        e.stopPropagation();
+        if (e.preventDefault) e.preventDefault();
+        e.returnValue = false;
+    }
 
-        guac_mouse.currentState.x = pageX - element.offsetLeft;
-        guac_mouse.currentState.y = pageY - element.offsetTop;
+    function moveMouse(clientX, clientY) {
+
+        guac_mouse.currentState.x = clientX - element.offsetLeft;
+        guac_mouse.currentState.y = clientY - element.offsetTop;
 
         // This is all JUST so we can get the mouse position within the element
         var parent = element.offsetParent;
-        while (parent) {
-            if (parent.offsetLeft && parent.offsetTop) {
-                guac_mouse.currentState.x -= parent.offsetLeft;
-                guac_mouse.currentState.y -= parent.offsetTop;
-            }
+        while (parent && !(parent === document.body)) {
+            guac_mouse.currentState.x -= parent.offsetLeft - parent.scrollLeft;
+            guac_mouse.currentState.y -= parent.offsetTop  - parent.scrollTop;
+
             parent = parent.offsetParent;
         }
 
+        // Offset by document scroll amount
+        var documentScrollLeft = document.body.scrollLeft || document.documentElement.scrollLeft;
+        var documentScrollTop = document.body.scrollTop || document.documentElement.scrollTop;
+
+        guac_mouse.currentState.x -= parent.offsetLeft - documentScrollLeft;
+        guac_mouse.currentState.y -= parent.offsetTop  - documentScrollTop;
+
         if (guac_mouse.onmousemove)
             guac_mouse.onmousemove(guac_mouse.currentState);
 
@@ -119,44 +149,53 @@ Guacamole.Mouse = function(element) {
 
 
     // Block context menu so right-click gets sent properly
-    element.oncontextmenu = function(e) {
-        return false;
-    };
+    element.addEventListener("contextmenu", function(e) {
+        cancelEvent(e);
+    }, false);
 
-    element.onmousemove = function(e) {
+    element.addEventListener("mousemove", function(e) {
 
         // Don't handle if we aren't supposed to
         if (gesture_in_progress) return;
 
-        e.stopPropagation();
+        cancelEvent(e);
 
-        moveMouse(e.pageX, e.pageY);
+        moveMouse(e.clientX, e.clientY);
 
-    };
+    }, false);
 
+    var touch_count = 0;
     var last_touch_x = 0;
     var last_touch_y = 0;
     var last_touch_time = 0;
     var pixels_moved = 0;
 
+    var touch_buttons = {
+        1: "left",
+        2: "right",
+        3: "middle"
+    };
+
     var gesture_in_progress = false;
     var click_release_timeout = null;
 
-    element.ontouchend = function(e) {
+    element.addEventListener("touchend", function(e) {
         
-        // If we're handling a gesture
-        if (gesture_in_progress) {
+        cancelEvent(e);
             
-            e.stopPropagation();
-            e.preventDefault();
+        // If we're handling a gesture AND this is the last touch
+        if (gesture_in_progress && e.touches.length == 0) {
             
             var time = new Date().getTime();
 
-            // If mouse already down, release anad clear timeout
-            if (guac_mouse.currentState.left) {
+            // Get corresponding mouse button
+            var button = touch_buttons[touch_count];
 
-                // Fire left button up event
-                guac_mouse.currentState.left = false;
+            // If mouse already down, release anad clear timeout
+            if (guac_mouse.currentState[button]) {
+
+                // Fire button up event
+                guac_mouse.currentState[button] = false;
                 if (guac_mouse.onmouseup)
                     guac_mouse.onmouseup(guac_mouse.currentState);
 
@@ -169,10 +208,11 @@ Guacamole.Mouse = function(element) {
             }
 
             // If single tap detected (based on time and distance)
-            if (time - last_touch_time <= 250 && pixels_moved < 10) {
+            if (time - last_touch_time <= guac_mouse.clickTimingThreshold
+                    && pixels_moved < guac_mouse.clickMoveThreshold) {
 
-                // Fire left button down event
-                guac_mouse.currentState.left = true;
+                // Fire button down event
+                guac_mouse.currentState[button] = true;
                 if (guac_mouse.onmousedown)
                     guac_mouse.onmousedown(guac_mouse.currentState);
 
@@ -180,72 +220,81 @@ Guacamole.Mouse = function(element) {
                 // touchstart within timeout.
                 click_release_timeout = window.setTimeout(function() {
                     
-                    // Fire left button up event
-                    guac_mouse.currentState.left = false;
+                    // Fire button up event
+                    guac_mouse.currentState[button] = false;
                     if (guac_mouse.onmouseup)
                         guac_mouse.onmouseup(guac_mouse.currentState);
                     
-                    // Allow mouse events now that touching is over
+                    // Gesture now over
                     gesture_in_progress = false;
-            
-                }, 250);
+
+                }, guac_mouse.clickTimingThreshold);
 
             }
 
+            // If we're not waiting to see if this is a click, stop gesture
+            if (!click_release_timeout)
+                gesture_in_progress = false;
+
         }
 
-    };
+    }, false);
 
-    element.ontouchstart = function(e) {
+    element.addEventListener("touchstart", function(e) {
 
-        // Record initial touch location and time for single-touch movement
+        cancelEvent(e);
+
+        // Track number of touches, but no more than three
+        touch_count = Math.min(e.touches.length, 3);
+
+        // Clear timeout, if set
+        if (click_release_timeout) {
+            window.clearTimeout(click_release_timeout);
+            click_release_timeout = null;
+        }
+
+        // Record initial touch location and time for touch movement
         // and tap gestures
-        if (e.touches.length == 1) {
-
-            e.stopPropagation();
-            e.preventDefault();
+        if (!gesture_in_progress) {
 
             // Stop mouse events while touching
             gesture_in_progress = true;
 
-            // Clear timeout, if set
-            if (click_release_timeout) {
-                window.clearTimeout(click_release_timeout);
-                click_release_timeout = null;
-            }
-
             // Record touch location and time
             var starting_touch = e.touches[0];
-            last_touch_x = starting_touch.pageX;
-            last_touch_y = starting_touch.pageY;
+            last_touch_x = starting_touch.clientX;
+            last_touch_y = starting_touch.clientY;
             last_touch_time = new Date().getTime();
             pixels_moved = 0;
 
-            // TODO: Handle different buttons
-
         }
 
-    };
+    }, false);
 
-    element.ontouchmove = function(e) {
+    element.addEventListener("touchmove", function(e) {
 
-        // Handle single-touch movement gesture (touchpad mouse move)
-        if (e.touches.length == 1) {
+        cancelEvent(e);
 
-            e.stopPropagation();
-            e.preventDefault();
+        // Get change in touch location
+        var touch = e.touches[0];
+        var delta_x = touch.clientX - last_touch_x;
+        var delta_y = touch.clientY - last_touch_y;
 
-            // Get change in touch location
-            var touch = e.touches[0];
-            var delta_x = touch.pageX - last_touch_x;
-            var delta_y = touch.pageY - last_touch_y;
+        // Track pixels moved
+        pixels_moved += Math.abs(delta_x) + Math.abs(delta_y);
 
-            // Track pixels moved
-            pixels_moved += Math.abs(delta_x) + Math.abs(delta_y);
+        // If only one touch involved, this is mouse move
+        if (touch_count == 1) {
+
+            // Calculate average velocity in Manhatten pixels per millisecond
+            var velocity = pixels_moved / (new Date().getTime() - last_touch_time);
+
+            // Scale mouse movement relative to velocity
+            var scale = 1 + velocity;
 
             // Update mouse location
-            guac_mouse.currentState.x += delta_x;
-            guac_mouse.currentState.y += delta_y;
+            guac_mouse.currentState.x += delta_x*scale;
+            guac_mouse.currentState.y += delta_y*scale;
 
             // Prevent mouse from leaving screen
 
@@ -264,20 +313,50 @@ Guacamole.Mouse = function(element) {
                 guac_mouse.onmousemove(guac_mouse.currentState);
 
             // Update touch location
-            last_touch_x = touch.pageX;
-            last_touch_y = touch.pageY;
+            last_touch_x = touch.clientX;
+            last_touch_y = touch.clientY;
 
         }
 
-    };
+        // Interpret two-finger swipe as scrollwheel
+        else if (touch_count == 2) {
+
+            // If change in location passes threshold for scroll
+            if (Math.abs(delta_y) >= guac_mouse.scrollThreshold) {
+
+                // Decide button based on Y movement direction
+                var button;
+                if (delta_y > 0) button = "down";
+                else             button = "up";
+
+                // Fire button down event
+                guac_mouse.currentState[button] = true;
+                if (guac_mouse.onmousedown)
+                    guac_mouse.onmousedown(guac_mouse.currentState);
+
+                // Fire button up event
+                guac_mouse.currentState[button] = false;
+                if (guac_mouse.onmouseup)
+                    guac_mouse.onmouseup(guac_mouse.currentState);
+
+                // Only update touch location after a scroll has been
+                // detected
+                last_touch_x = touch.clientX;
+                last_touch_y = touch.clientY;
+
+            }
+
+        }
+
+    }, false);
 
 
-    element.onmousedown = function(e) {
+    element.addEventListener("mousedown", function(e) {
 
         // Don't handle if we aren't supposed to
         if (gesture_in_progress) return;
 
-        e.stopPropagation();
+        cancelEvent(e);
 
         switch (e.button) {
             case 0:
@@ -294,15 +373,15 @@ Guacamole.Mouse = function(element) {
         if (guac_mouse.onmousedown)
             guac_mouse.onmousedown(guac_mouse.currentState);
 
-    };
+    }, false);
 
 
-    element.onmouseup = function(e) {
+    element.addEventListener("mouseup", function(e) {
 
         // Don't handle if we aren't supposed to
         if (gesture_in_progress) return;
 
-        e.stopPropagation();
+        cancelEvent(e);
 
         switch (e.button) {
             case 0:
@@ -319,14 +398,25 @@ Guacamole.Mouse = function(element) {
         if (guac_mouse.onmouseup)
             guac_mouse.onmouseup(guac_mouse.currentState);
 
-    };
+    }, false);
 
-    element.onmouseout = function(e) {
+    element.addEventListener("mouseout", function(e) {
 
         // Don't handle if we aren't supposed to
         if (gesture_in_progress) return;
 
-        e.stopPropagation();
+        // Get parent of the element the mouse pointer is leaving
+       	if (!e) e = window.event;
+
+        // Check that mouseout is due to actually LEAVING the element
+        var target = e.relatedTarget || e.toElement;
+        while (target != null) {
+            if (target === element)
+                return;
+            target = target.parentNode;
+        }
+
+        cancelEvent(e);
 
         // Release all buttons
         if (guac_mouse.currentState.left
@@ -341,15 +431,15 @@ Guacamole.Mouse = function(element) {
                 guac_mouse.onmouseup(guac_mouse.currentState);
         }
 
-    };
+    }, false);
 
     // Override selection on mouse event element.
-    element.onselectstart = function() {
-        return false;
-    };
+    element.addEventListener("selectstart", function(e) {
+        cancelEvent(e);
+    }, false);
 
     // Scroll wheel support
-    element.onmousewheel = function(e) {
+    function mousewheel_handler(e) {
 
         // Don't handle if we aren't supposed to
         if (gesture_in_progress) return;
@@ -386,14 +476,11 @@ Guacamole.Mouse = function(element) {
             }
         }
 
-        if (e.preventDefault)
-            e.preventDefault();
+        cancelEvent(e);
 
-        e.returnValue = false;
-
-    };
-
-    element.addEventListener('DOMMouseScroll', element.onmousewheel, false);
+    }
+    element.addEventListener('DOMMouseScroll', mousewheel_handler, false);
+    element.addEventListener('mousewheel',     mousewheel_handler, false);
 
 };
 
